
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <libssp.h>

#include <wiringPi.h>

#include <signal.h>

#include<unistd.h>

#include<opencv/cv.h>
#include <opencv/highgui.h>
// sudo apt-get install libopencv-dev
// sudo apt-get install libgl1-mesa-dri

#include <sys/time.h>

#include<pthread.h>

#include <zbar.h>
zbar_image_scanner_t *scanner = NULL;

float font_scale = 1.0;
int line_break = 10;

int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

static int frame_count;

volatile int stop_process;
static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

    printf("Frame dropped by camera.\n");

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}

IplImage* BayerImg, *RGBImg;
IplImage* imGray;

#define WaitCount 200

void GammaCorrection(IplImage* src, IplImage* dst, double gamma){
  int i;
  uchar LUT[256];

  for (i = 0; i < 256; i++){
    LUT[i] = (int)(pow((double)i / 255.0, 1.0 / gamma) * 255.0);
  }
  CvMat lut_mat = cvMat(1, 256, CV_8UC1, LUT);
  cvLUT(src, dst, &lut_mat);
}

volatile int SettingMode = 0;
#define GAIN_MODE 0
#define SS_MODE 1 // Shutter Speed
volatile int GammaEnable = 0;

struct __paracopy_args {
  int width;
  int height;
  unsigned char *src;
  unsigned char *dest;
};

void *paracopy_thread_func(void *param){
  struct __paracopy_args *copyparams = (struct __paracopy_args *)param;
  int i;
  for(i=0;i<copyparams->height;i++)
    memcpy(copyparams->dest+copyparams->width*i,
	   copyparams->src+copyparams->width*i,
	   copyparams->width);

  return NULL;
}

char *model_name;

#define GAIN_CHANGE_VAL 25

#define NUM_COPY_THREAD 4
pthread_t thread_handle[NUM_COPY_THREAD];
struct __paracopy_args copy_params[NUM_COPY_THREAD];

void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  //  printf("Preprocess is active. frame_count = %d\n", frame_count);
  //if(frame_count % 10 == 0) printf("frame_count = %d\n", frame_count);
  
  int i,j;
  int Width = ssp_get_frame_width(frame);
  int Height = ssp_get_frame_height(frame);
  unsigned char *frame_data = ssp_get_frame_data(frame);
#if 0 // Simple copy
  for(i=0;i<BayerImg->height;i++)
    for(j=0;j<BayerImg->width;j++){
      ((unsigned char *)(BayerImg->imageData + BayerImg->widthStep * i))[j] = frame_data[i*BayerImg->widthStep+j];
    }
#endif

#if 0 // memcpy
  for(i=0;i<BayerImg->height;i++){
    memcpy((unsigned char *)(BayerImg->imageData + BayerImg->widthStep * i), &frame_data[i*BayerImg->widthStep], BayerImg->width);
  }
#endif
#if 0 // long long copy
  for(i=0;i<BayerImg->height;i++)
    for(j=0;j<BayerImg->width/sizeof(long long);j++){
      *((long long *)(BayerImg->imageData + BayerImg->widthStep * i) + j) = *((long long*)(frame_data+i*BayerImg->widthStep)+j);
    }
#endif

#if 1 // parallel copy
  for(i=0;i<NUM_COPY_THREAD;i++){
    copy_params[i].width = BayerImg->widthStep;
    copy_params[i].height = BayerImg->height/NUM_COPY_THREAD;
    copy_params[i].dest = (unsigned char *)(BayerImg->imageData + BayerImg->widthStep * (i*BayerImg->height/NUM_COPY_THREAD));
    copy_params[i].src = &frame_data[(i*BayerImg->height/NUM_COPY_THREAD)*BayerImg->widthStep];
    pthread_create(&(thread_handle[i]) , NULL , paracopy_thread_func ,&(copy_params[i]));
      
  }
  for(i=0;i<NUM_COPY_THREAD;i++)
    pthread_join(thread_handle[i] , NULL);
#endif
  
    ssp_release_frame(frame);

    cvCvtColor( BayerImg, RGBImg, CV_BayerRG2RGB);
    if(GammaEnable == 1)
      GammaCorrection(RGBImg, RGBImg, 2.2);

    // barcode detection
    cvCvtColor(RGBImg,imGray, CV_RGB2GRAY);

    zbar_image_t *image = zbar_image_create();
    char format[] = "Y800";
    unsigned long fourcc = ((format[0] & 0xff) |
			    ((format[1] & 0xff) << 8) |
			    ((format[2] & 0xff) << 16) |
			    ((format[3] & 0xff) << 24));
    zbar_image_set_format(image, fourcc);
    zbar_image_set_size(image, imGray->widthStep, imGray->height);
    //    zbar_image_set_data(image, imGray->imageData, imGray->widthStep*imGray->height, zbar_image_free_data);
    zbar_image_set_data(image, imGray->imageData, imGray->widthStep*imGray->height, NULL);

    int n = zbar_scan_image(scanner, image);

#if 1
    CvFont font[1];
    CvScalar color={0,255,0,0};
    
    cvInitFont(&font[0], CV_FONT_HERSHEY_SIMPLEX, font_scale,font_scale, 0, 1, 8);

    char msgTxt[1024];
    const zbar_symbol_t *symbol = zbar_image_first_symbol(image);
    int vplace = 30;
    for(; symbol; symbol = zbar_symbol_next(symbol)) {
      zbar_symbol_type_t typ = zbar_symbol_get_type(symbol);
      const char *data = zbar_symbol_get_data(symbol);
      int nlocation = zbar_symbol_get_loc_size(symbol);
      int *barcodex, *barcodey;

      barcodex = (int *)malloc(sizeof(int)*nlocation);
      barcodey = (int *)malloc(sizeof(int)*nlocation);
      for(i=0;i<nlocation;i++){
	barcodex[i] = zbar_symbol_get_loc_x(symbol, i);
	barcodey[i] = zbar_symbol_get_loc_y(symbol, i);
	printf("(%d,%d)\n", barcodex[i], barcodey[i]);
      }


      int lefttopx, lefttopy;
      int leftbottomx, leftbottomy;
      int righttopx, righttopy;
      int rightbottomx, rightbottomy;
      lefttopx = barcodex[0] < barcodex[1] ? barcodex[0] : barcodex[1];
      lefttopx = lefttopx < barcodex[2] ? lefttopx : barcodex[2];
      rightbottomx = barcodex[0] > barcodex[1] ? barcodex[0] : barcodex[1];
      rightbottomx = rightbottomx > barcodex[2] ? rightbottomx : barcodex[2];
      lefttopy = barcodey[0] < barcodey[1] ? barcodey[0] : barcodey[1];
      lefttopy = lefttopy < barcodey[2] ? lefttopy : barcodey[2];
      rightbottomy = barcodey[0] > barcodey[1] ? barcodey[0] : barcodey[1];
      rightbottomy = rightbottomy > barcodey[2] ? rightbottomy : barcodey[2];
      
      //printf("(%d,%d)-(%d,%d)\n", lefttopx, lefttopy, rightbottomx,rightbottomy);

      // find min x location
      int min, max;
      int curr_i = 0;
      min = 10000000;
      for(i=0;i<nlocation;i++)
	if(min > barcodex[i]){
	  min = barcodex[i];
	  curr_i = i;
	}
      lefttopx = min;
      //      lefttopy = barcodey[curr_i];

      min = 10000000;
      for(i=0;i<nlocation;i++)
	if(min > barcodey[i])
	  min = barcodey[i];
	  lefttopy = min;
      max = 0;
      curr_i = 0;
      for(i=0;i<nlocation;i++)
	if(max < barcodex[i]){
	  max = barcodex[i];
	  curr_i = i;
	}
      rightbottomx = max;
      //      rightbottomy = barcodey[curr_i];
      max = 0;
      for(i=0;i<nlocation;i++)
	if(max < barcodey[i])
	  max = barcodey[i];
	  rightbottomy = max;
      
      /*      if(nlocation > 1){
	for(i=0;i<nlocation;i++){
	  if(i == (nlocation-1))
	    cvLine(RGBImg, cvPoint(barcodex[nlocation-1], barcodey[nlocation-1]),
		   cvPoint(barcodex[0], barcodey[0]), color, 3, 8, 0);
	  else
	    cvLine(RGBImg, cvPoint(barcodex[i], barcodey[i]),
		   cvPoint(barcodex[i+1], barcodey[i+1]), color, 3, 8, 0);
	}
	}*/

      cvRectangle(RGBImg, cvPoint(lefttopx, lefttopy),
	     cvPoint(rightbottomx, rightbottomy), color, 3, 8, 0);
      
      free(barcodex);
      free(barcodey);
      
      printf("decoded %s:\n%s\n",
	     zbar_get_symbol_name(typ), data);

      sprintf(msgTxt, "decoded %s:",
	      zbar_get_symbol_name(typ));
      cvPutText (RGBImg, msgTxt, cvPoint (15,vplace), &font[0], color);
      vplace += (int)(50*font_scale);
      
      int data_length = strlen(data);
      curr_i = 0;
      while(data_length > 0){
	int len = line_break>data_length? data_length: line_break;
	printf("data_length = %d\n", data_length);
	for(i=0;i<len; i++, curr_i++)
	  msgTxt[i] = data[curr_i];
	msgTxt[i] = '\0';
	//puts(msgTxt);
	cvPutText (RGBImg, msgTxt, cvPoint (15,vplace), &font[0], color);
	vplace += (int)(50*font_scale);
	data_length -= len;
      }
    }

#endif
    
    zbar_image_destroy(image);

    
    cvShowImage("SSP Viewer", RGBImg);
    //cvShowImage("SSP Viewer", imGray);
    int KeyCode = cvWaitKey(2);
    KeyCode &= 0xFF;
    if(KeyCode == 'q'){
      stop_process=1;
    }
    if(KeyCode == 's'){
      printf("Setting Mode is set to Shutter Speed Mode.\n");
      SettingMode = SS_MODE;
    }
    if(KeyCode == 'g'){
      printf("Setting Mode is set to Gain Mode.\n");
      SettingMode = GAIN_MODE;
    }
    if(KeyCode == '0'){
      if(GammaEnable == 0){
	GammaEnable = 1;
	printf("Gamma Correction is Enabled.\n");
      }
      else{
	GammaEnable = 0;
	printf("Gamma Correction is Disabled.\n");
      }
    }
    if(KeyCode == '+' || KeyCode == '-'){
      int RegVal;

      if(SettingMode == GAIN_MODE){
	if(strcmp(model_name, "IMX397") == 0){
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal > 224) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 232;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name, "IMX219") == 0){
	  if(__CCIRegRead(handle, 0x0157, &RegVal) == 0)
	    printf("Reading register error.\n");
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 233) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 232;
	  }
	  if(__CCIRegWrite(handle, 0x0157, RegVal) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x0157, &RegVal) == 0)
	    printf("Reading register error.\n");
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name, "IU233") == 0){
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 225) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 224;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name, "IMX408") == 0){
	  int GainVal;
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x1) << 8);
	  RegVal = GainVal;
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 481) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 480;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal & 0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x0204, (RegVal>>8)&0x1) == 0)
	    printf("Writing register error.\n");	
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x1) << 8);
	  RegVal = GainVal;
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name, "IMX378") == 0){
	  int GainVal;
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x3) << 8);
	  RegVal = GainVal;
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 978) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 978;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal & 0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x0204, (RegVal>>8)&0x3) == 0)
	    printf("Writing register error.\n");	
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x3) << 8);
	  RegVal = GainVal;
	  printf("Analog Gain Set to %d\n", RegVal);

	}
      }

      if(SettingMode == SS_MODE){
	if(strcmp(model_name, "IMX219") == 0){
	  int frame_length_lines;
	  int coarse_integration_time;
	  
	  if(__CCIRegRead(handle, 0x0160, &RegVal) == 0)
	    printf("Reading register error.\n");
	  frame_length_lines = RegVal<<8;
	  if(__CCIRegRead(handle, 0x0161, &RegVal) == 0)
	    printf("Reading register error.\n");
	  frame_length_lines |= RegVal;

	  if(__CCIRegRead(handle, 0x015A, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time = RegVal<<8;
	  if(__CCIRegRead(handle, 0x015B, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time |= RegVal;

	  if(KeyCode == '+'){
	    coarse_integration_time ++;
	    if(coarse_integration_time == (frame_length_lines-5)) RegVal = 1;
	  }
	  if(KeyCode == '-'){
	    coarse_integration_time --;
	    if(coarse_integration_time == -1) coarse_integration_time = frame_length_lines-4;
	  }
	  if(__CCIRegWrite(handle, 0x015A, (coarse_integration_time>>8)&0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x015B, coarse_integration_time&0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x015A, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time = RegVal<<8;
	  if(__CCIRegRead(handle, 0x015B, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time |= RegVal;
	  printf("Shutter Speed Set (coarse_integration_time) to %d\n", coarse_integration_time);
	}
      }
    }
    
    //    printf("frame  count = %d\n", frame_count);
    frame_count ++;
}

int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE
int main(int argc, char **argv){

  int opt;

  char *profile_filename = NULL;
  char *filename = NULL;
  
  model_name = NULL;

  int full_screen = 0;
  
  while ((opt = getopt(argc, argv, "p:m:s:l:f")) != -1) {
    switch (opt) {
    case 'p':
      if(optarg == NULL){
	printf("Error: Specify profile XML file.\n");
	break;
      }
      printf("Profile: %s\n",optarg);
      profile_filename = (char *)malloc(strlen(optarg)+1);
      memset(profile_filename, 0, strlen(optarg)+1);
      strncpy(profile_filename, optarg, strlen(optarg)+1);
      break;
    case 'm':
      if(optarg == NULL){
	printf("Error: Specify sensor model name: IMX219 or IMX408 or IU233 or IMX378 or IMX397.\n");
	return -1;
      }
      printf("Model: %s\n",optarg);
      model_name = (char *)malloc(strlen(optarg)+1);
      memset(model_name, 0, strlen(optarg)+1);
      strncpy(model_name, optarg, strlen(optarg)+1);
      break;
    case 's':
      font_scale = atof(optarg);
      break;
    case 'l':
      line_break = atoi(optarg);
      break;
    case 'f':
      full_screen = 1;
      break;
    }      
  }

  //  puts(profile_filename);
  //puts(filename);
  
  if(profile_filename == NULL || model_name == NULL){
    printf("SSP library barcode sample program\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
    printf("Usage: %s -p profile_file -m model [-f -s font_size -l line_break]\n", argv[0]);
    printf("\tmodel: IMX219 or IMX408 or IU233 or IMX378 or IMX397\n");
    printf("\t-f: full screen mode\n");
    printf("\t-s font_scale: scale for non-zero value\n");
    printf("\t-l line_break: number of charactors for line break\n");
    return 0;
  }

  

  printf("Font Scale: %f\n",font_scale);
  printf("Line Break: %d\n",line_break);
  
  //  profile_filename = argv[1];
  //filename = argv[2];
  
  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;

  struct ssp_settings settings;

  frame_count = 0;
  
  settings.lib_settings.NumFrameFIFOSize = 2;
  settings.lib_settings.NumThreadForBuildInPreprocess = 4;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER8;
  settings.frame_preprocess_options.GammaCorrectionEnable = GAMMA_OPTION;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;

  struct ssp_handle *handle;

  struct ssp_profile *profile;
  
  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
	 profile->ImageProperty.Width,
	 profile->ImageProperty.Height,
	 profile->ImageProperty.Frequency,
	 profile->ImageProperty.BayerBits);


  // OpenCV Settings
  //  cvNamedWindow("SSP Viewer", CV_WINDOW_AUTOSIZE); 
  if(full_screen){
    cvNamedWindow("SSP Viewer", CV_WINDOW_KEEPRATIO);
    cvSetWindowProperty("SSP Viewer", CV_WND_PROP_FULLSCREEN, CV_WINDOW_FULLSCREEN);
  }
  else
    cvNamedWindow("SSP Viewer", CV_WINDOW_NORMAL | CV_WINDOW_KEEPRATIO); 

  BayerImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);
  RGBImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,3);
  imGray = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);


  // Setting Up barcode reader
  scanner = zbar_image_scanner_create();
  zbar_image_scanner_set_config(scanner, 0, ZBAR_CFG_ENABLE, 1);
  
  int i,j;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return 0;
    }
    printf("OK\n");
     
    printf("Initialize done.\n");
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }

    struct timeval tv_start, tv_end;
    struct timezone tz_local;
    gettimeofday(&tv_start, &tz_local);
    
    cvWaitKey(2);

    while(1){ if(stop_process==1) break;}
      
    gettimeofday(&tv_end, &tz_local);

    double start_time = tv_start.tv_sec+(double)tv_start.tv_usec/1000000;
    double end_time = tv_end.tv_sec+(double)tv_end.tv_usec/1000000;
    double fps;
    fps = (double)frame_count/(end_time-start_time);
    printf("FrameRate = %f\n", fps);
    
    printf("FrameCount = %d\n", frame_count);
    printf("stop streaming\n");
    if((retval = ssp_stop_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }
    
    printf("Flush Wait.\n");
    ssp_flush_event(handle);
    printf("Flush Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
      return 0;
    }
    
    cvReleaseImage(&RGBImg);
    cvReleaseImage(&BayerImg);
    cvDestroyWindow("SSP Viewer");
    
    zbar_image_scanner_destroy(scanner);    

    printf("Finalize done.\n");fflush(stdout);
  }
  return 1;
}
